@@ -6,6 +6,9 @@
 
 /*
  * command.c - maintain lists of commands
+ *
+ * 01/20/00 (seiwald) - Upgraded from K&R to ANSI C
+ * 09/08/00 (seiwald) - bulletproof PIECEMEAL size computation
  */
 
 # include "jam.h"
@@ -26,22 +29,24 @@ cmd_new(
 	RULE	*rule,
 	LIST	*targets,
 	LIST	*sources,
-	LIST	*shell )
+	LIST	*shell,
+	int	maxline )
 {
 	CMD *cmd = (CMD *)malloc( sizeof( CMD ) );
 
 	cmd->rule = rule;
 	cmd->shell = shell;
 	cmd->next = 0;
+	cmd->buf = (char *)malloc( maxline );
 
 	lol_init( &cmd->args );
 	lol_add( &cmd->args, targets );
 	lol_add( &cmd->args, sources );
 
-	/* Bail if the result won't fit in MAXLINE */
+	/* Bail if the result won't fit in maxline */
 	/* We don't free targets/sources/shell if bailing. */
 
-	if( var_string( rule->actions, cmd->buf, MAXLINE, &cmd->args ) < 0 )
+	if( var_string( rule->actions, cmd->buf, maxline, &cmd->args ) < 0 )
 	{
 	    cmd_free( cmd );
 	    return 0;
@@ -59,5 +64,6 @@ cmd_free( CMD *cmd )
 {
 	lol_free( &cmd->args );
 	list_free( cmd->shell );
+	free( (char *)cmd->buf );
 	free( (char *)cmd );
 }
