@@ -6,6 +6,8 @@
 
 /*
  * pathsys.h - PATHNAME struct 
+ *
+ * 11/04/02 (seiwald) - const-ing for string literals
  */
 
 /*
@@ -24,12 +26,13 @@ typedef struct _pathname PATHNAME;
 typedef struct _pathpart PATHPART;
 
 struct _pathpart {
-	char	*ptr;
+	const char *ptr;
 	int	len;
 };
 
 struct _pathname {
 	PATHPART	part[6];
+	char		delim;
 # ifdef OS_VMS
 	int		parent;
 # endif
@@ -44,6 +47,6 @@ struct _pathname {
 } ;
 
 void path_build( PATHNAME *f, char *file, int binding );
-void path_parse( char *file, PATHNAME *f );
+void path_parse( const char *file, PATHNAME *f );
 void path_parent( PATHNAME *f );
 
